/**
 * Copyright (C) 2011  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.reporting.server;

import com.google.gwt.user.server.rpc.RemoteServiceServlet;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Properties;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;
import javax.servlet.ServletContext;
import javax.servlet.ServletException;
import org.eclipse.birt.core.exception.BirtException;
import org.eclipse.birt.report.engine.api.EngineException;
import org.eclipse.birt.report.engine.api.HTMLRenderOption;
import org.eclipse.birt.report.engine.api.HTMLServerImageHandler;
import org.eclipse.birt.report.engine.api.IReportEngine;
import org.eclipse.birt.report.engine.api.IReportRunnable;
import org.eclipse.birt.report.engine.api.IRunAndRenderTask;
import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;
import org.ow2.orchestra.reporting.client.WebReport;
import org.ow2.orchestra.util.Misc;


/**
 * This servlet is responsible of handling the generation of a report using the
 * Birt engine runtime version 3.7.0
 */
public class WebReportImpl extends RemoteServiceServlet implements WebReport {

  private static final long serialVersionUID = 1L;
  public static final String DESIGN_SUFFIX = ".rptdesign";

  /**
   * Birt engine initialization
   */
  private IReportEngine birtReportEngine = null;

  private Map<String, IReportRunnable> reportRunnables = new HashMap<String, IReportRunnable>();
  /**
   * Orchestra Connexion properties
   */
  private String orchestraServiceUrl;
  private String orchestraObjectName;


  private ServletContext sc;

  /**
   * logger used to log out the exceptions
   */
  protected static final Logger LOG = Logger.getLogger(WebReportImpl.class.getName());

  public WebReportImpl() throws IOException {
    super();
    HashMap<String, String> orchestraValues = retrieveOrchestraValues();
    orchestraServiceUrl = orchestraValues.get("ORCHESTRA_URL");
    orchestraObjectName = orchestraValues.get("ORCHESTRA_OBJECTNAME");

  }

  public String getReport(String reportName) {

    ByteArrayOutputStream myarray = new ByteArrayOutputStream();

    /**
     * openReport Design( ) methods in the ReportEngine class is to load
     * the report what is an xml file , and this returns IReportRunnable
     * instance that represents the engine's view of the report design
     */
    Misc.fastDynamicLog(LOG, Level.FINEST, "opening report %s", reportName);
    IReportRunnable design = reportRunnables.get(reportName);
    if (design == null) {
      throw new OrchestraRuntimeException("Report " + reportName + " not found.");
    }
    Misc.fastDynamicLog(LOG, Level.FINEST, "report opened, creating task");
    // create task to run and render the report
    IRunAndRenderTask task = birtReportEngine.createRunAndRenderTask(design);
    Misc.fastDynamicLog(LOG, Level.FINEST, "task created");
    // to set Birt's parameter's value according
    task.setParameterValue("PARAMBIRT_ORCHESTRA_URL", orchestraServiceUrl);
    task.setParameterValue("PARAMBIRT_ORCHESTRA_OBJECTNAME", orchestraObjectName);
    // set output options
    HTMLRenderOption options = new HTMLRenderOption();
    // options.setOutputFormat(format);
    options.setOutputFormat(HTMLRenderOption.OUTPUT_FORMAT_HTML);
    // using the base image url
    options.setSupportedImageFormats("PNG;GIF;JPG;BMP;SVG");
    //The engine will write images here
    String tmpDir = File.separator + "birt-temp" + File.separator + "images";
    //TODO currently images are stored within the exploded war
    // Those images should be stored into a temporary directory
    // outside the war.
    options.setImageDirectory(sc.getRealPath(tmpDir));
    options.setBaseImageURL(sc.getContextPath() + tmpDir);
    options.setImageHandler(new HTMLServerImageHandler());
    options.setEmbeddable(true);
    // end image section
    options.setOutputStream(myarray);
    // to load the file with options
    task.setRenderOption(options);
    // to run the report
    Misc.fastDynamicLog(LOG, Level.FINEST, "running generation");
    try {
      task.run();
    } catch (EngineException e) {
      throw new OrchestraRuntimeException("Exception during report generation", e);
    }
    Misc.fastDynamicLog(LOG, Level.FINEST, "generation complete");

    return myarray.toString();

  }

  /**
   * Initialization of the servlet. The servlet's init method calls the
   * BirtEngine.initBirtConfig() method to read the BIRT configuration file
   * discussed earlier. *
   *
   * @throws ServletException if an error occure
   */
  public void init() throws ServletException {
    super.init();
    BirtEngine.initBirtConfig();
    sc = getServletContext();
    try {
      /**
       *  If it's the second time we are calling this servlet to get a second report
       *  it means that  BirtEngine.getBirtEngine(sc) ! = null so it will not initialize the Birt engine again
       */
      this.birtReportEngine = BirtEngine.getBirtEngine(sc);
    } catch (BirtException e) {
      throw new ServletException(e);
    }
    // Find reports in the WAR
    Map<String, InputStream> reports = findReports();
    // Pre load report to speed up first generation.
    for (Entry<String, InputStream> report : reports.entrySet()) {
      Misc.fastDynamicLog(LOG, Level.FINEST, "pre-loading report: " + report.getKey());
      try {
        IReportRunnable irr = birtReportEngine.openReportDesign(report.getValue());
        this.reportRunnables.put(report.getKey(), irr);
      } catch (EngineException e) {
        throw new ServletException(e);
      }
    }
  }

  /**
   * Find all "*.rptdesign" resource in the War. Open an input stream for each report.
   *
   * @return
   * @throws ServletException
   */
  private Map<String, InputStream> findReports() throws ServletException {
    Map<String, InputStream> reports = new HashMap<String, InputStream>();
    Set<String> resources = sc.getResourcePaths("/WEB-INF/classes");
    for (String resource : resources) {
      if (resource.endsWith(DESIGN_SUFFIX)) {
        reports.put(resource.substring("/WEB-INF/classes".length()), sc.getResourceAsStream(resource));
      }
    }
    try {
      Set<String> resourceJars = sc.getResourcePaths("/WEB-INF/lib");
      for (String resourceJar : resourceJars) {
        if (resourceJar.endsWith(".jar")) {
          Misc.fastDynamicLog(LOG, Level.FINEST, "checking: " + resourceJar);
          InputStream jarInputStream = sc.getResourceAsStream(resourceJar);
          ZipInputStream zipInputStream = new ZipInputStream(jarInputStream);
          ZipEntry zipEntry = null;
          while ((zipEntry = zipInputStream.getNextEntry()) != null) {
            if (!zipEntry.isDirectory()) {
              if (zipEntry.getName().endsWith(DESIGN_SUFFIX)) {
                Misc.fastDynamicLog(LOG, Level.FINEST, "adding: " + zipEntry.getName());
                final ByteArrayOutputStream baos = new ByteArrayOutputStream();
                int c;
                final byte[] buffer = new byte[512];
                while ((c = zipInputStream.read(buffer)) != -1) {
                  baos.write(buffer, 0, c);
                }
                baos.flush();
                reports.put(zipEntry.getName(), new ByteArrayInputStream(baos.toByteArray()));
                Misc.close(baos);
              }
            }
          }
          zipInputStream.close();
          jarInputStream.close();
        }
      }
    } catch (IOException e) {
      throw new ServletException(e);
    }
    return reports;
  }

  /**
   * Destruction of the servlet, however we do not need destroying the
   * instance each time as the initialization takes few time.
   */
  public void destroy() {
    BirtEngine.destroyBirtEngine();
    super.destroy();

  }

  /**
   * Method to set the map's values 'url service' and 'object Name' from
   * orchestra.properties written on the console side.
   *
   * @return Returns config map
   * @throws java.io.IOException
   */

  private HashMap<String, String> retrieveOrchestraValues() throws IOException {

    HashMap<String, String> myMap = new HashMap<String, String>();
    InputStream inputStream = null;

    try {
      final File file = new File("conf" + File.separator + "orchestra.properties");
      final Properties properties = new Properties();
      if (file.exists()) {
        inputStream = new FileInputStream(file);
      } else {
        inputStream = Thread.currentThread().getContextClassLoader().getResourceAsStream("orchestra.properties");
      }
      properties.load(inputStream);
      orchestraServiceUrl = properties.getProperty("orchestra.jmx.serviceUrl");
      orchestraObjectName = properties.getProperty("orchestra.jmx.objectName");
      myMap.put("ORCHESTRA_URL", orchestraServiceUrl);
      myMap.put("ORCHESTRA_OBJECTNAME", orchestraObjectName);

    } finally {
      Misc.close(inputStream);
    }

    return myMap;
  }

}
